class Sign < ActiveRecord::Base 
  belongs_to :trail
  belongs_to :guidepost
  delegate :osm_trail, :kct_trail, to: :trail
  delegate :kctref, :nazev, :has_kctref?, :title, :osm_guidepost, to: :guidepost
  
  # asociované objekty  
  def osm_signs
    raise "Neexistuje OSM trail s osmid: #{self.trail.osmid}, ačkoli trail: #{self.trail.inspect} se na něj odkazuje" if osm_trail.nil?
    return [] if self.guidepost.osmid.nil?
    raise "Neexistuje OSM guidepost s osmid #{self.guidepost.osmid.inspect}, ačkoli guidepost dbf:#{guidepost.id} se na něj odkazuje" if osm_guidepost.nil?
    r = OsmSign.where(:osm_guidepost_id => osm_guidepost.id, :osm_trail_id => osm_trail.id)
  end
  def osm_sign; osm_signs.first end  
  def kct_sign
    raise "Neexistuje KCT trail s kctref: #{self.trail.kctref}, která ale patří k trail:   #{self.trail.inspect} nemá" if kct_trail.nil?
    raise "Neexistuje KCT guidepost s kctref #{self.guidepost.kctref}, který ale patří k guidepostu dbf id:#{guidepost.id}" if kct_guidepost.nil?
    KctKilometraz.where( :kct_trail => kct_trail, :kct_guidepost => kct_guidepost ).first
  end
  def kct_guidepost
    raise self.inspect if guidepost.nil?
    guidepost.kct_guidepost
  end
    
  # Vyhleda a pripadne vytvori Sign dle zadanych parametru. Priklady
  # Sign.create_from_or_update(kctref: 'PS012', osmid: 1234567, trail: 142)
  # Sign.create_from_or_update(kct_kilometraz: KctKilometraz.find(11))
  # Sign.create_from_or_update(kct_kilometraz: KctKilometraz.find(11), trail: 142)  # umožňuje vykopírovat informace z KctKilometraze aniz by byla trail k kct_trail asociovana  
  # Sign.create_from_or_update(guidepost: Guidepost.find(157), trail: 142)  
  def self.create_from_or_update(opts)
    kctref, osmid, trail, kct_kilometraz, guidepost_id = opts[:kctref], opts[:osmid], opts[:trail], opts[:kct_kilometraz], opts[:guidepost]
    
    # vytěžit kct_kilometraz pokud je nastaveno
    if not kct_kilometraz.nil?
      kctref = kct_kilometraz.kctref unless kctref
      trail  = kct_kilometraz.kct_trail.trail unless trail
    end
    
    # find or create Guidepost
    guidepost_by_kctref = Guidepost.find_by_kctref(kctref)
    guidepost_by_kctref_id = guidepost_by_kctref.andand.id
    guidepost_by_osmid = osmid ? Guidepost.find_by_osmid(osmid) : nil
    guidepost_by_osmid_id = guidepost_by_osmid.andand.id
    guidepost_by_guidepost_id = guidepost_id.is_a?(Guidepost) ? guidepost_id : guidepost_id.nil? ? nil : Guidepost.find(guidepost_id)
    guidepost_by_guidepost_id_id = guidepost_by_guidepost_id.andand.id
    
    foundids = [guidepost_by_kctref_id,guidepost_by_osmid_id, guidepost_by_guidepost_id].uniq.compact
    
    case foundids.size
      when 0 then # create
        guidepost = Guidepost.create(:kctref => kctref, :osmid => osmid )
      when 1 then 
        guidepost = Guidepost.find(foundids.first)
      else
        raise "Options identify more then one guidepost:
          kctref:'#{kctref.inspect}' gives Guidepost##{guidepost_by_kctref_id},  
          osmid:#{guidepost_by_osmid.andand.osmid || 'N/A'} gives Guidepost##{guidepost_by_osmid_id || 'N/A'} and 
          guidepost_id:#{guidepost_by_guidepost_id_id} gives Guidepost##{guidepost_by_guidepost_id}. 
          Can not resolve the conflict."
    end
    
    # fill in data
    guidepost.kctref = kctref unless kctref.nil? or guidepost.has_kctref?(kctref)
    guidepost.osmid = osmid unless osmid.nil?
    guidepost.nazev = kct_kilometraz.kct_guidepost.nazev  if kct_kilometraz and guidepost.nazev.blank?
    
    if guidepost.save
      # find or create sign
      sign = trail.signs.where(:guidepost => guidepost).first
      if sign.nil?
        sign = Sign.create(:trail=>trail, :guidepost=>guidepost)
      end
      if kct_kilometraz
        sign.km = kct_kilometraz.km
        sign.save!
      end
    else
      raise "There was an error saving guidepost (#{guidepost.inspect}): #{guidepost.errors.full_messages.join}. Called with opts= #{opts.inspect}"
    end
    sign
  end
  
  # returns number which denotes (not normalized) probability that these signs are identical and can be merged
  def self.merge_score(kct_sign,osm_sign_or_guidepost,osm_trail=nil)
    merge_reasons_and_score(kct_sign,osm_sign_or_guidepost,osm_trail)[1]
  end
  # returns textual reasons for merge
  def self.merge_reasons(kct_sign,osm_sign_or_guidepost,osm_trail=nil)
    merge_reasons_and_score(kct_sign,osm_sign_or_guidepost,osm_trail)[0]
  end
  
  def self.merge_reasons_and_score(kct_sign,osm_sign_or_guidepost,osm_trail=nil)
    return ['',0] if osm_sign_or_guidepost.nil? or kct_sign.nil?
    if osm_sign_or_guidepost.is_a?(OsmGuidepost)
      @osm_guidepost = osm_sign_or_guidepost
      @osm_sign = nil
    else
      @osm_guidepost = osm_sign_or_guidepost.osm_guidepost
      @osm_sign = osm_sign_or_guidepost
    end
    @kct_guidepost = kct_sign.kct_guidepost
    score = 0; reasons=[]
   
    # by references and associations
    if @kct_guidepost.kctref == @osm_guidepost.kctref
      score += 100
      reasons << 'equal kctrefs'
    end
    if (@kct_guidepost.guidepost == @osm_guidepost.guidepost) and (!@kct_guidepost.guidepost.nil?)
      score += 150
      reasons << 'Guidepost ifself is identical (maybe due to other trail)'
    end
    
    # by names
    levenshtein_distance = Text::Levenshtein.distance(@kct_guidepost.nazev_normalized,@osm_guidepost.nazev_normalized || '')
    score +=
      case levenshtein_distance
        when 0 then 70
        when 1 then 55
        when 2 then 25
        else 0
      end
    reasons<< "Levenshtein distance of names is small (#{levenshtein_distance})"  if levenshtein_distance<=2
      
    # by position
    if @osm_sign.nil?
      estimated_kct_km = osm_trail.nil? ? nil : osm_trail.kilometraz_of_nearest_point_to(@osm_guidepost.coordinates)
    else
      estimated_kct_km = @osm_sign.estimated_km_from_kct_kms
    end
    unless estimated_kct_km.nil?
      score += 
        case (estimated_kct_km-kct_sign.km).abs
          when 0 .. 0.1 then 50
          when 0.1 .. 0.3 then 40
          when 0.3 .. 0.5 then 35
          when 0.5 .. 0.75 then 25
          when 0.75 .. 0.99 then 15
          when 0.99 .. 1.5 then 5
          when 1.5 .. 2 then 1
          else 0
        end
      reasons << "Difference of positions on trails is small (#{'%.0f' % ((estimated_kct_km-kct_sign.km).abs*1000)}m)" if (estimated_kct_km-kct_sign.km).abs<1.5      
    end
    
    # when osm_guidepost is NOT on osm_trail and distance to osm_trail is large, it is very bad sign
    if @osm_sign.nil? and !osm_trail.nil?
      dist = osm_trail.distance_to(@osm_guidepost)
      scoreadj =
        case dist
          when 0..30 then 0
          when 30..60 then -30
          when 60..100 then -50
          when 100..300 then -150
          when 300..600 then -200
          when 600..1000 then -250
          else -300
        end
      score += scoreadj
      reasons << "Bad sign distance to osm_trail large (#{dist} m => #{scoreadj}pt) " if (scoreadj < -60)       
    end
      
    return reasons,score
  end
  
  def update_km_from_osm!
    self.km = if not osm_sign.nil?
      self.osm_sign.km
    elsif not osm_guidepost.nil?
      self.trail.osm_trail.gpx.kilometraz_of_nearest_point_to(osm_guidepost.coordinates)
    else
      self.km
    end
  end
    
  # hash obsahující řádky textu (objekty SignLine). jako key je textový label typu "2. řádek dolní směrovky"
  def lines
    if text2.blank?
      SignLine.sign_lines_from_text(text1)
    else
      SignLine.sign_lines_from_text(text1, 'horní směrovky').merge(
        SignLine.sign_lines_from_text(text2, 'dolní směrovky')
      )
    end
  end
  
  def problems
    lines.sum ([]) do |line_origin, line|
      line.problems.collect{ |problem| (problem.gsub /\?/, "#{line_origin} \"#{line.text}\"").html_safe }
    end
  end
  
  ## DECORATORS  viz cells/sign_cell.rb
end

class SignLine
  attr_reader :text, :km, :nazev, :problems, :type
  
  def initialize(atext)
    @text = atext; 
    @problems = []; @km = nil; @nazev = nil
    if /\s*(?<nazev>.*)\s+(?<km>\d+((\.|,)\d+)?)\s*(?<unit>km|)\s*$/ =~ @text
      @nazev = $~[:nazev]
      @km = $~[:km].to_f
      @unit = $~[:unit]
      @problems << '? nemá jednotky. Předpokládám km, ale prosím doplňte.' if @unit != 'km'
      @type = :regular
    elsif /^dále po .*$/ =~ @text
      @type = :dale
    else
      @problems << '? mi nedává smysl'
      @type = :unrecognizes
    end
  end

  def regular?; @type == :regular end
  def dale?; @type == :dale end
    
  # vyrobí hash který má popisky v keys a SignLine objects ve values
  def self.sign_lines_from_text(text, smerovka= '')
    smerovka = ' '+smerovka unless smerovka.blank?
    res = {}
    text.to_s.split(/\r?\n/).each_with_index do |line,index|
      sline = SignLine.new(line)
      break if sline.dale? # pokud je v textu "dále po ..." je třeba další řádky ignorovat
      res["#{index+1}. řádek#{smerovka}"] = sline 
    end
    return res
  end
end

